#define CGAL_EIGEN3_ENABLED

#include <maya/MGlobal.h>
#include <maya/MDataBlock.h>
#include <maya/MDataHandle.h>
#include <maya/MPlug.h>
#include <maya/MObject.h>
#include <maya/MTypeId.h>
#include <maya/MFnMesh.h>
#include <maya/MFnMeshData.h>
#include <maya/MFloatPointArray.h>
#include <maya/MPointArray.h>
#include <maya/MFnGenericAttribute.h>
#include <maya/MFnNumericAttribute.h>
#include <maya/MString.h>
#include <maya/MStringArray.h>

#include "skeletonize.h"
#include "../pluginMain.h"

#include <CGAL/Simple_cartesian.h>
#include <CGAL/Surface_mesh.h>
#include <CGAL/boost/graph/graph_traits_Surface_mesh.h>
#include <CGAL/boost/graph/properties_Surface_mesh.h>
#include <CGAL/Mean_curvature_flow_skeletonization.h>
#include <fstream>
typedef CGAL::Simple_cartesian<double>                           K;
typedef K::Point_3                                               Point;
typedef CGAL::Surface_mesh<Point>                                Triangle_mesh;
typedef boost::graph_traits<Triangle_mesh>::vertex_descriptor    vertex_descriptor;
typedef CGAL::Mean_curvature_flow_skeletonization<Triangle_mesh> Skeletonization;
typedef Skeletonization::Skeleton                                Skeleton;
typedef Skeleton::vertex_descriptor                              Skeleton_vertex;
typedef Skeleton::edge_descriptor                                Skeleton_edge;

MTypeId skeletonize::id(ID_SKELETONIZE);
MObject skeletonize::aInMesh;
MObject skeletonize::aOutMesh;
MObject skeletonize::aQuads;
MObject skeletonize::aOffset;

skeletonize::skeletonize(){}
skeletonize::~skeletonize(){}

void* skeletonize::creator()
{
	return new skeletonize();
}

MStatus skeletonize::initialize()
{
	MFnGenericAttribute gAttr;
	MFnNumericAttribute nAttr;

	aInMesh = gAttr.create("inMesh", "im");
	gAttr.addDataAccept(MFnData::kMesh);
	gAttr.setStorable(false);
	gAttr.setDisconnectBehavior(MFnAttribute::kReset);

	aOutMesh = gAttr.create("outMesh", "om");
	gAttr.addDataAccept(MFnData::kMesh);
	gAttr.setWritable(false);
	gAttr.setStorable(false);

	aQuads = nAttr.create("quads", "qds", MFnNumericData::kBoolean, true);
	nAttr.setKeyable(true);

	aOffset = nAttr.create("offset", "off", MFnNumericData::kFloat, 0.001);
	nAttr.setKeyable(true);

	addAttribute(aInMesh);
	addAttribute(aOutMesh);
	addAttribute(aQuads);
	addAttribute(aOffset);

	attributeAffects(aInMesh, aOutMesh);
	attributeAffects(aQuads, aOutMesh);
	attributeAffects(aOffset, aOutMesh);

	return MS::kSuccess;
}

MStatus skeletonize::compute(const MPlug& plug, MDataBlock& data)
{
	if(data.outputValue(state).asShort() == 2) return MS::kUnknownParameter;

	if(plug != aOutMesh) return MS::kUnknownParameter;

	MFnMeshData md;
	MObject o = md.create();
	MFnMesh m;
	m.copy(data.inputValue(aInMesh).asMesh(), o);
	m.setObject(o);

	unsigned int pntcnt = m.numVertices();
	MIntArray cnts, conn;
	m.getTriangles(cnts, conn);
	unsigned int tricnt = cnts.length();
	if(pntcnt > 2 && tricnt > 1)
	{
		MFloatPointArray pnts;
		m.getPoints(pnts, MSpace::kWorld);

		Triangle_mesh mesh;
		std::vector<vertex_descriptor> vd(pntcnt);
		for(unsigned int i = 0; i < pntcnt; ++i) vd[i] = mesh.add_vertex(K::Point_3(pnts[i].x, pnts[i].y, pnts[i].z));
		for(unsigned int i = 0; i < tricnt; ++i) mesh.add_face(vd[conn[i*3]], vd[conn[i*3+1]], vd[conn[i*3+2]]);
		Skeleton skeleton;
		Skeletonization mcs(mesh);
		mcs.contract_until_convergence();
		mcs.convert_to_skeleton(skeleton);

		cout<<boost::num_vertices(skeleton)<<endl;
		if(boost::num_vertices(skeleton) > 2)
		{
			bool quads = data.inputValue(aQuads).asBool();
			float offset = data.inputValue(aOffset).asFloat();
			if(offset > -0.0001 && offset <= 0) offset = -0.0001;
			else if(offset >= 0 && offset < 0.0001) offset = 0.0001;
			MPointArray pnts2;
			if(quads) pnts2.setLength(4);
			else pnts2.setLength(3);
			Point p;
			MFnMesh m2;
			m2.create(0, 0, MFloatPointArray(), MIntArray(), MIntArray(), o);
			BOOST_FOREACH(Skeleton_edge e, edges(skeleton))
			{
				if(quads)
				{
					p = skeleton[source(e, skeleton)].point;
					pnts2[3].x = pnts2[0].x = p[0];
					pnts2[3].y = pnts2[0].y = p[1];
					pnts2[3].z = pnts2[0].z = p[2];
					pnts2[3].x += offset;
					p = skeleton[target(e, skeleton)].point;
					pnts2[2].x = pnts2[1].x = p[0];
					pnts2[2].y = pnts2[1].y = p[1];
					pnts2[2].z = pnts2[1].z = p[2];
					pnts2[2].x += offset;
					m2.addPolygon(pnts2);
				}
				else
				{
					p = skeleton[source(e, skeleton)].point;
					pnts2[0].x = p[0];
					pnts2[0].y = p[1];
					pnts2[0].z = p[2];
					p = skeleton[target(e, skeleton)].point;
					pnts2[1].x = p[0];
					pnts2[1].y = p[1];
					pnts2[1].z = p[2];
					pnts2[2] = (pnts2[0] + pnts2[1]) * 0.5;
					m2.addPolygon(pnts2);
				}
			}
		}
		else
		{
			MStringArray sa;
			MPlug(thisMObject(), aInMesh).name().split('.', sa);
			MGlobal::displayWarning(sa[0]+": Cannot proceed - input mesh has hole(s).");
		}
	}

	data.outputValue(aOutMesh).set(o);

	return MS::kSuccess;
}
